// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdio.h>

#include "bfs.h"
#include "util.h"

void init_dist(int_board dist)
{
	int i, j;
	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		dist[i][j] = 1000000;
}

void bfs_from_place(board b, int_board dist, int start_y, int start_x,
		int ignore_boxes)
{
	// do bfs, assuming "dist" is initialized

	int current_x[MAX_SIZE * MAX_SIZE];
	int current_y[MAX_SIZE * MAX_SIZE];
	int next_x[MAX_SIZE * MAX_SIZE];
	int next_y[MAX_SIZE * MAX_SIZE];

	int current_size, next_size;
	int i, j, y, x, d;

	int occupied_mask = OCCUPIED;

	if (ignore_boxes)
		occupied_mask = WALL;

	current_x[0] = start_x;
	current_y[0] = start_y;
	current_size = 1;

	d = 0;
	dist[start_y][start_x] = d;

	do
	{
		next_size = 0;
		for (i = 0; i < current_size; i++)
		{
			for (j = 0; j < 4; j++)
			{
				x = current_x[i] + delta_x[j];
				y = current_y[i] + delta_y[j];

				if (((b[y][x] & occupied_mask) == 0) && (dist[y][x] == 1000000))
				{
					next_x[next_size] = x;
					next_y[next_size] = y;
					dist[y][x] = d + 1;
					next_size++;
				}
			}
		} // i

		d++;

		for (i = 0; i < next_size; i++)
		{
			current_x[i] = next_x[i];
			current_y[i] = next_y[i];
		}
		current_size = next_size;

	} while (current_size > 0);
}


void bfs(board b, int_board dist, int start_y, int start_x,
	int ignore_boxes)
{
	init_dist(dist);

	bfs_from_place(b, dist,  start_y,  start_x, ignore_boxes);
}
 

int mark_connectivities(board b, int_board dist)
{
	int_board local_dist;
	int c = 0;
	int y, x, i, j;

	init_dist(dist);

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
	{
		if (inner[i][j] == 0) continue;
		if (dist[i][j] != 1000000) continue;
		if (b[i][j] & OCCUPIED) continue;

		init_dist(local_dist);
		bfs_from_place(b, local_dist, i, j, 0);

		for (y = 0; y < height; y++)
		for (x = 0;  x < width; x++)
		if (local_dist[y][x] < 1000000)
			dist[y][x] = c;

		c++;
	}
	return c;
}

void mark_connectivities_around_place(board b, int_board dist,
									 int start_y, int start_x)
{
	int_board local_dist;
	int c, i, j, y, x;

	init_dist(dist);

	for (c = 0; c < 4 ; c++)
	{
		i = start_y + delta_y[c];
		j = start_x + delta_x[c];

		if (inner[i][j] == 0) continue;
		if (dist[i][j] != 1000000) continue;
		if (b[i][j] & OCCUPIED) continue;

		init_dist(local_dist);
		bfs_from_place(b, local_dist, i, j, 0);

		for (y = 0; y < height; y++)
			for (x = 0; x < width; x++)
				if (local_dist[y][x] < 1000000)
					dist[y][x] = c;
	}
}
