// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#ifndef __BOARD
#define __BOARD

#include "global.h"

typedef unsigned char board[MAX_SIZE][MAX_SIZE];
typedef int int_board[MAX_SIZE][MAX_SIZE];

extern board initial_board;

void copy_board(board from, board to);

void init_inner(board b);

extern int index_num;
extern board inner;

void init_index_x_y();
int y_x_to_index(int y, int x);
void index_to_y_x(int index, int *y, int *x);
void clear_boxes(board b, board board_without_boxes);
int board_contains_boxes(board b);
void turn_boxes_into_walls(board b);

void expand_sokoban_cloud(board b);
void get_sokoban_position(board b, int *y, int *x);
void clear_sokoban_inplace(board b);
void clear_targets_inplace(board b);
int board_is_solved(board b);


UINT_64 get_board_hash(board b);
void board_to_bytes(board b, UINT_8 *data);
void bytes_to_board(UINT_8 *data, board b);

void print_board(board b);
void save_initial_board(board b);

int boxes_in_level(board b);

void zero_board(board b);

void keep_boxes_in_inner(board b);

int get_indices_of_boxes(board b, int *indices);
int get_indices_of_targets(board b, int *indices);
int get_indices_of_bases(board b, int *indices);
void clear_bases_inplace(board b);

#endif