// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdlib.h>

#include "global.h"
#include "challenges.h"
#include "curry.h"
#include "util.h"

int N;

board *challenges[MAX_BOXES + 1][200];
int board_solved[MAX_BOXES + 1][200];
int board_difficulty[MAX_BOXES + 1][200];

void set_one_challenge(int i, int j)
{
	challenges[i][j] = (board*)malloc(sizeof(board));
	if (challenges[i][j] == 0) exit_with_error("can't allocate challenge");

	setup_board_with_k_boxes(*(challenges[i][j]), i);
	board_solved[i][j] = 0;
	board_difficulty[i][j] = 0;
}

void init_challenges(int box_num, int episodes_num)
{
	int i, j;

	for (i = 1; i <= box_num; i++)
		for (j = 0; j < episodes_num; j++)
			set_one_challenge(i, j);

	N = episodes_num;
}

void free_challenges(int box_num)
{
	int i, j;

	for (i = 1; i <= box_num; i++)
		for (j = 0; j < N; j++)
			free(challenges[i][j]);
}

void replace_challenge(int box_num, int index)
{
	free(challenges[box_num][index]);

	set_one_challenge(box_num, index);
}


void get_challenge(int box_num, int index, board b)
{
	copy_board(*(challenges[box_num][index]), b);
}

int is_board_solved(int box_num, int index)
{
	return board_solved[box_num][index];
}


void mark_board_as_solved(int box_num, int index)
{
	board_solved[box_num][index] = 1;
}

int solved_all_challenges(int k)
{
	int i;

	for (i = 0; i < N; i++)
		if (board_solved[k][i] == 0)
			return 0;
	return 1;
}

void update_board_difficulty(int box_num, int index, int difficulty)
{
	board_difficulty[box_num][index] += difficulty;
}

int get_board_difficulty(int box_num, int index)
{
	return board_difficulty[box_num][index];
}
