// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdio.h>

#include "graph.h"
#include "distance.h"
#include "util.h"

int distance_from_to[MAX_INNER][MAX_INNER];

int can_get_to_dest[MAX_INNER];
int can_get_from_start[MAX_INNER];

int impossible_place[MAX_INNER];

void compute_can_get_to_dest(board b)
{
	int i, x, y, place;

	for (i = 0; i < index_num; i++)
	{
		can_get_to_dest[i] = 0;

		for (y = 0; y < height; y++)
		{
			for (x = 0; x < width; x++)
			{
				if (inner[y][x] == 0) continue;

				if (b[y][x] & TARGET)
				{
					place = y_x_to_index(y, x);
					if (distance_from_to[i][place] < 1000000)
						can_get_to_dest[i] = 1;
				}
			}
		}
	}
}


void compute_can_from_start(board b)
{
	int i, x, y, place;

	for (i = 0; i < index_num; i++)
	{
		can_get_from_start[i] = 0;

		for (y = 0; y < height; y++)
		{
			for (x = 0; x < width; x++)
			{
				if (inner[y][x] == 0) continue;

				if (b[y][x] & BOX)
				{
					place = y_x_to_index(y, x);
					if (distance_from_to[place][i] < 1000000)
						can_get_from_start[i] = 1;
				}
			}
		}
	}
}

void set_impossible_places()
{
	int i;

	for (i = 0; i < MAX_INNER; i++)
		impossible_place[i] = 1;

	for (i = 0; i < index_num; i++)
	if ((can_get_from_start[i]) && (can_get_to_dest[i]))
		impossible_place[i] = 0;

}


void set_distances(board b)
{
	board board_without_boxes;

	int from_index, to_index;
	int dist;

	clear_boxes(b, board_without_boxes);

	build_graph(board_without_boxes);

	for (from_index = 0; from_index < index_num; from_index++)
	{
		set_graph_weights_to_infinity();
		clear_weight_around_cell(from_index);
		do_graph_iterations();
		
		for (to_index = 0; to_index < index_num; to_index++)
		{
			dist = get_weight_around_cell(to_index);
			distance_from_to[from_index][to_index] = dist;
		}
	}

	compute_can_get_to_dest(b);
	compute_can_from_start(b);
	set_impossible_places();

}

