// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdlib.h>

#include "util.h"
#include "episode.h"

void allocate_episodes(episodes *es, int n)
{
	es->e = (episode*)malloc(sizeof(episode) * n);
	if (es->e == 0) exit_with_error("can't alloc episodes");
	es->episodes_num = 0;
}


void free_episodes(episodes *es)
{
	free(es->e);
	es->episodes_num = 0;
}

void setup_board_from_episode(board b, episode *e)
{
	int i, y, x;

	clear_boxes(initial_board, b);
	clear_targets_inplace(b);
	clear_sokoban_inplace(b);

	for (i = 0; i < e->box_num; i++)
	{
		index_to_y_x(e->bases[i], &y, &x);
		b[y][x] |= BOX;
		b[y][x] |= BASE;

		index_to_y_x(e->targets[i], &y, &x);
		b[y][x] |= TARGET;
	}

	b[global_initial_player_y][global_initial_player_x] |= SOKOBAN;

	expand_sokoban_cloud(b);
}



void print_episode(episode *e)
{
	board b;
	int i;

	move *moves;

	moves = e->moves;

	setup_board_from_episode(b, e);
	print_board(b);
	for (i = 0; i < e->len; i++)
	{
		apply_move(b, moves + i);
		print_board(b);
	}
}


// SIMBA
void get_suggestion_from_episode(int *b_has_box, int *b_has_target,
	episode *e, move *m, int *value)
{
	int active_indices[MAX_INNER];
	int indices_with_boxes[MAX_INNER];

	int i, t,from, to;
	int overlap = 0;
	int best_overlap;
	int best_place = 0;
	move *moves;

	// find indices with different targets in b and episodes
	
	for (i = 0; i < index_num; i++)
		active_indices[i] = b_has_target[i] ^ e->has_target[i] ^ 1;
	
	// place boxes		
	for (i = 0; i < index_num; i++)
		indices_with_boxes[i] = e->has_base[i];


	// init overlap
	for (i = 0; i < index_num; i++)
		overlap += active_indices[i] & indices_with_boxes[i] & b_has_box[i];

	best_overlap = overlap;

	moves = e->moves;

	// go through moves and see if overlap improves
	for (t = 0; t < e->len; t++)
	{
		from = moves[t].from;
		to   = moves[t].to;

		overlap -= b_has_box[from] & active_indices[from];

		indices_with_boxes[from] = 0;
		indices_with_boxes[to] = 1;

		overlap += b_has_box[to] & active_indices[to];

		if (overlap > best_overlap)
		{
			best_overlap = overlap;
			best_place = t + 1;
		}
	}

	*value = (e->len - best_place);

	best_place++; 
	if (best_place >= e->len)
		best_place = e->len;

	*m = moves[best_place - 1];
}

int is_easier_episode(episode *new_e, episode *old_e)
{
	if (new_e->box_num < old_e->box_num) return 1;
	if (new_e->box_num > old_e->box_num) return 0;

	if (new_e->difficulty < old_e->difficulty) return 1;
	if (new_e->difficulty > old_e->difficulty) return 0;

	return 0;
}

int find_easiest_episode(episodes *es)
{
	int i;
	int best_place = 0;

	for (i = 1; i < es->episodes_num; i++)
	{
		if (is_easier_episode(es->e + i, es->e + best_place))
			best_place = i;
	}

	return best_place;
}
