// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include "fixed_boxes.h"
#include "util.h"

int turn_2x2_to_walls(board b)
{
	// this routine does not guarantee to blockify all boxes in one pass
	int i, j, y, x;
	int sum;
	int changed = 0;

	for (i = 0; i < (height - 1); i++)
	for (j = 0; j < (width - 1); j++)
	{
		sum = 0;

		for (y = 0; y < 2; y++)
		for (x = 0; x < 2; x++)
		if (b[i + y][j + x] & OCCUPIED) sum++;

		if (sum == 4)
		{
			// check if one of them is a box
			for (y = 0; y < 2; y++)
				for (x = 0; x < 2; x++)
				{
					if (b[i + y][j + x] & BOX) 
						changed = 1;
					b[i + y][j + x] = WALL;
				}
		}
	}
	return changed;
}

int close_holes(board b)
{
	int i, j, k;

	for (i = 1; i < (height - 1); i++)
	{
		for (j = 1; j < (width - 1); j++)
		{
			if (b[i][j] == SPACE)
			{
				// close holes of type
				//    $
				//   # #
				//    #

				for (k = 0; k < 4; k++)
				{
					if (b[i + delta_y[(k + 1) & 3]][j + delta_x[(k + 1) & 3]] != WALL) continue;
					if (b[i + delta_y[(k + 2) & 3]][j + delta_x[(k + 2) & 3]] != WALL) continue;
					if (b[i + delta_y[(k + 3) & 3]][j + delta_x[(k + 3) & 3]] != WALL) continue;
					b[i][j] = WALL;
					return 1;
				}
			}

			if (b[i][j] == SOKOBAN)
			{
				// close holes of type
				//    .
				//   #.#
				//    #

				for (k = 0; k < 4; k++)
				{
					if ((b[i + delta_y[(k + 0) & 3]][j + delta_x[(k + 0) & 3]] & SOKOBAN) == 0) continue;
					if (b[i + delta_y[(k + 1) & 3]][j + delta_x[(k + 1) & 3]] != WALL) continue;
					if (b[i + delta_y[(k + 2) & 3]][j + delta_x[(k + 2) & 3]] != WALL) continue;
					if (b[i + delta_y[(k + 3) & 3]][j + delta_x[(k + 3) & 3]] != WALL) continue;
					b[i][j] = WALL;
					return 1;
				}
			}
		}
	}

	return 0;
}


void turn_fixed_boxes_into_walls(board b, board w)
{
	int changed = 1;

	copy_board(b, w);	

	do
	{
		changed = turn_2x2_to_walls(w);
		changed |= close_holes(w);
	} 
	while (changed);
}

