// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham


#include "stdafx.h"

#include <stdio.h>

#include "match_deadlock.h"
#include "util.h"
#include "distance.h"
#include "hungarian.h"


int check_matching_deadlock(board b)
{
	int box_places[MAX_INNER];
	int box_num = 0;
	int target_places[MAX_INNER];
	int target_num = 0;
	int i, j, y, x, from, to;

	static int cost[MAX_BOXES][MAX_BOXES];
	hungarian_solution sol;
	
	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);
		if (b[y][x] & BOX)
			box_places[box_num++] = i;
		if (b[y][x] & TARGET)
			target_places[target_num++] = i;
	}

	if (box_num != target_num)
		return 1;

	if (box_num >= MAX_BOXES)
		exit_with_error("hungarian size overflow");

	for (i = 0; i < box_num; i++)
		for (j = 0; j < box_num; j++)
			cost[i][j] = 1;

	for (i = 0; i < box_num; i++)
	{
		for (j = 0; j < box_num; j++)
		{
			from = box_places[i];
			to = target_places[j];

			if (distance_from_to[from][to] != 1000000)
				cost[i][j] = 0;
		}
	}
	
	solve_hungarian(box_num, cost, &sol);

	if (sol.weight == 0)
		return 0;

	return 1;
}
