// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include "preprocess.h"


int sanity_checks(board b)
{
	int box_num = boxes_in_level(b);

	if (box_num >= MAX_BOXES)
	{
		printf("%d boxes\n", box_num);
		printf("too many boxes\n");
		sprintf(global_fail_reason, "Too many boxes");
		return 0;
	}
	return 1;
}

void close_holes_in_board(board b)
{
	int changed = 0;
	int i, j, k, sum;

	do
	{
		changed = 0;

		for (i = 1; i < (height - 1); i++)
			for (j = 1; j < (width - 1); j++)
			{
				if (b[i][j] != 0) continue;

				sum = 0;
				for (k = 0; k < 4; k++)
					if (b[i + delta_y[k]][j + delta_x[k]] == WALL)
						sum++;
				if (sum >= 3)
				{
					b[i][j] = WALL;
					changed = 1;
				}
			}
	} while (changed);

}


void remove_boxes_out_of_inner(board b)
{
	board inner;
	int i, j;

	// mark inner and boxes in inner
	clear_boxes(b, inner);
	expand_sokoban_cloud(inner);

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if ((inner[i][j] & SOKOBAN) == 0)
			{
				b[i][j] &= ~BOX;
				b[i][j] &= ~TARGET;
			}
}

void turn_decorative_boxes_to_walls(board b)
{
	board c;
	int i, j, k;
	int changed = 1;

	remove_boxes_out_of_inner(b);

	clear_boxes(b, c);

	// eliminate pullable boxes

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (c[i][j] & TARGET)
				c[i][j] = BOX;

	while (changed)
	{
		changed = 0;

		for (i = 0; i < height; i++)
			for (j = 0; j < width; j++)
			{
				if ((c[i][j] & BOX) == 0) continue;
				for (k = 0; k < 4; k++)
				{
					if (c[i + delta_y[k]][j + delta_x[k]] & OCCUPIED) continue;
					if (c[i + delta_y[k] * 2][j + delta_x[k] * 2] & OCCUPIED) continue;

					c[i][j] &= ~BOX;
					changed = 1;
				}
			}
	}

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (c[i][j] & BOX)
				b[i][j] = WALL;



	// eliminate pushable boxes
	copy_board(b, c);
	changed = 1;

	while (changed)
	{
		changed = 0;

		for (i = 0; i < height; i++)
			for (j = 0; j < width; j++)
			{
				if ((c[i][j] & BOX) == 0) continue;
				for (k = 0; k < 4; k++)
				{
					if (c[i + delta_y[k]][j + delta_x[k]] & OCCUPIED) continue;
					if (c[i - delta_y[k]][j - delta_x[k]] & OCCUPIED) continue;

					c[i][j] &= ~BOX;
					changed = 1;
				}
			}
	}

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (c[i][j] & BOX)
				b[i][j] = WALL;
}