// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdlib.h>

#include "select.h"
#include "util.h"
#include "tree.h"

int inv_epsilon = 20;


int son_visited_in_walk(tree *t, expansion_data *e, int son)
{
	node_element *n;
	extern int walk_id;

	n = get_son(t, e, son);
	if (n->expansion == -1)
		return 0;

	e = t->expansions + n->expansion;
	if (e->walk_id == walk_id)
		return 1;

	return 0;
}


int choose_random_son(tree *t, expansion_data *e)
{
	int son;
	move_hash_data *mh = t->move_hashes + e->move_hash_place;
	int active_sons[MAX_MOVES];
	int active_sons_num = 0;
	int i,node;

	if (e->node->deadlocked) exit_with_error("choosing son for deadlocked position");

	for (i = 0; i < e->moves_num; i++)
	{
		if (mh[i].deadlocked) continue;
		if (son_visited_in_walk(t, e, i)) continue;

		active_sons[active_sons_num++] = i;
	}

	if (active_sons_num == 0)
		return -1;

	son = active_sons[rand() % active_sons_num];
	node = find_node_by_hash(t, mh[son].hash);
	if (node == -1) exit_with_error("missing node");

	if (t->nodes[node].deadlocked)
		exit_with_error("not marked as deadlock2");

	return son;
}



int choose_by_probability(tree *t, expansion_data *e)
{

	int i;
	move_hash_data *mh;
	float prob[MAX_MOVES];
	int deadlocked = 0;
	int visited = 0;

	if (e->node->deadlocked) exit_with_error("deadlocked node");

	mh = t->move_hashes + e->move_hash_place;

	for (i = 0; i < e->moves_num; i++)
	{
		if (mh[i].deadlocked)
		{
			prob[i] = 0;
			deadlocked++;
			continue;
		}

		if (son_visited_in_walk(t, e, i))
		{
			prob[i] = 0;
			visited++;
			continue;
		}

		prob[i] = mh[i].weight;
		
	}

	if ((deadlocked + visited) == e->moves_num)
	{
		if (visited == 0) exit_with_error("this is strange");
		return -1;
	}


	normalize_vector(prob, e->moves_num);

	i = choose_with_probability(prob, e->moves_num);

	if (prob[i] == 0) exit_with_error("prob zero");

	return i;
}


int select_son(tree *t, expansion_data *e)
{
	int son;

	if (e->node->deadlocked) exit_with_error("node deadlocked");

	if (has_a_winning_son(t, e, &son))
		return son;

	set_deadlock_status(t, e);
	if (e->node->deadlocked) return -1;

	if ((rand() % inv_epsilon) == 0)
	{
		son = choose_random_son(t,e);
		return son;
	}

	son = choose_by_probability(t, e);
	if (son == -1) return -1;

	return son;
}
