// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "global.h"
#include "board.h"
#include "distance.h"
#include "deadlock.h"
#include "io.h"
#include "sol.h"
#include "preprocess.h"
#include "curry.h"
#include "tree.h"
#include "engine.h"

int preprocess_level(board b)
{
	strcpy(global_fail_reason, "Unknown reason");

	if (sanity_checks(b) == 0)
		return 0;

	turn_decorative_boxes_to_walls(b);
	close_holes_in_board(b);

	init_inner(b);
	init_index_x_y();

	keep_boxes_in_inner(b);
	save_initial_board(b);
	expand_sokoban_cloud(b);

	set_forbidden_tunnel();

	if (verbose >= 3)
	{
		printf("\nLevel %d:\n", level_id);
		print_board(b);
	}

	set_distances(b);

	return 1;
}



void solve_level_set()
{	
	board b;
	int i,solved;
	int levels_num, res;
	int from, to;

	levels_num = load_level_from_file(b, 10000); // count levels in file

	write_log_header();
	write_solution_header();

	from = 1;
	to = levels_num;

	if (global_from_level != -1) from = global_from_level;
	if (global_to_level   != -1) to   = global_to_level;
	if (just_one_level != -1)
		from = to = just_one_level;

	for (i = from; i <= to; i++) // using one-based here
	{
		solved = 0;

		level_id = i;
		printf("\n=================\nLevel %d\n=================\n", level_id);

		res = load_level_from_file(b, level_id);

		save_level_to_solution_file(b);

		start_time = (int)time(0);

		res = preprocess_level(b);
		if (res)
			solved = curry_engine(b);		

		end_time = (int)time(0);

		save_level_log(solved);
		save_times_to_solutions_file(end_time - start_time);
	}

	write_log_footer();
}


void process_args(int argc, char **argv)
{
	int i;

	if (global_dir[0] == '.') // not experimental
	{
		if (argc < 2)
		{
			printf("Usage: curry.exe levels.sok [options]");
			exit(0);
		}

		strcpy(global_level_set_name, argv[1]);
	}
	

	for (i = 0; i < argc; i++)
	{
		if (strcmp(argv[i], "-level") == 0)
			sscanf(argv[i + 1], "%d", &just_one_level);

		if (strcmp(argv[i], "-out_dir") == 0)
			strcpy(global_dir, argv[i + 1]);

		if (strcmp(argv[i], "-time") == 0)
			sscanf(argv[i + 1], "%d", &time_limit);

		if (strcmp(argv[i], "-from") == 0)
			sscanf(argv[i + 1], "%d", &global_from_level);

		if (strcmp(argv[i], "-to") == 0)
			sscanf(argv[i + 1], "%d", &global_to_level);

		if (strcmp(argv[i], "-seed") == 0)
			sscanf(argv[i + 1], "%d", &seed);

		if (strcmp(argv[i], "-subcases") == 0)
			sscanf(argv[i + 1], "%d", &subcases_num);

	}
}


int main(int argc, char **argv)
{
	process_args(argc, argv);

	printf("==============================\n");
	printf("%s\n", solver_name);
	printf("Copyright 2021 by Yaron Shoham\n");
	printf("==============================\n");

	init_search_tree();
	allocate_past_episodes(subcases_num);
 
	solve_level_set();  

	return 0;
}

