// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <string.h>

#include "util.h"
#include "sol.h"
#include "io.h"
#include "util.h"
#include "moves.h"

int find_sol_move(board b_in,
	int start_sok_index, int start_box_index, 
	int end_sok_index, int end_box_index,
	char *moves)
{
	int space_size;
	space_size = index_num * index_num;
	int *src;
	int i, j;
	static int current[MAX_INNER * 4]; // I'm not sure about the bound here...
	static int next[MAX_INNER * 4];
	int current_num = 0;
	int next_num = 0;
	int sok_index, box_index, sok_y, sok_x, box_y, box_x;
	int next_sok_y, next_sok_x, next_box_y, next_box_x;
	int current_index, next_index;
	char m;
	int moves_num = 0;
	board b;

	copy_board(b_in, b);
	index_to_y_x(start_box_index, &box_y, &box_x);
	if ((b[box_y][box_x] & BOX) == 0) 
		exit_with_error("missing box1");
	b[box_y][box_x] &= ~BOX;

	src = (int*)malloc(space_size * sizeof(int));
	if (src == NULL)
	{
		exit_with_error("could not alloc sol");
	}

	for (i = 0; i < space_size; i++)
		src[i] = -1;

	current[0] = end_sok_index * index_num + end_box_index; // we'll start from the end and pull back
	src[current[0]] = current[0];
	current_num = 1;

	while (current_num)
	{
		next_num = 0;

		for (i = 0; i < current_num; i++)
		{
			current_index = current[i];

			sok_index = current_index / index_num;
			box_index = current_index % index_num;

			index_to_y_x(sok_index, &sok_y, &sok_x);
			index_to_y_x(box_index, &box_y, &box_x);

			// try to move the player
			for (j = 0; j < 4; j++)
			{
				next_sok_y = sok_y + delta_y[j];
				next_sok_x = sok_x + delta_x[j];

				if ((next_sok_y == box_y) && (next_sok_x == box_x))
					continue; // that's a push, deal with it later

				if (b[next_sok_y][next_sok_x] & OCCUPIED) continue;

				next_index = y_x_to_index(next_sok_y, next_sok_x) * index_num + box_index;
				if (src[next_index] != -1) continue;

				src[next_index] = current_index;
				next[next_num++] = next_index;
			}

			// try to pull box
			for (j = 0; j < 4; j++)
			{
				next_box_y = box_y + delta_y[j];
				next_box_x = box_x + delta_x[j];

				if (next_box_y != sok_y) continue;
				if (next_box_x != sok_x) continue;

				next_sok_y = sok_y + delta_y[j];
				next_sok_x = sok_x + delta_x[j];

				if (b[next_sok_y][next_sok_x] & OCCUPIED) continue;

				next_index = y_x_to_index(next_sok_y, next_sok_x) * index_num +
					y_x_to_index(next_box_y, next_box_x);

				if (src[next_index] != -1) continue;

				src[next_index] = current_index;
				next[next_num++] = next_index;
			}

			if (next_num >= (MAX_INNER * 4))
				exit_with_error("next_num overflow");
		} // loop on current

		current_num = 0;
		for (i = 0; i < next_num; i++)
			current[current_num++] = next[i];
	} // while current_num

	// now display the path

	current_index = start_sok_index * index_num + start_box_index;

	if (src[current_index] == -1)
		exit_with_error("could not find path");

	next_index = src[current_index];

	while (next_index != current_index)
	{
		index_to_y_x(current_index / index_num, &sok_y, &sok_x);

		index_to_y_x(next_index / index_num, &next_sok_y, &next_sok_x);
		index_to_y_x(next_index % index_num, &next_box_y, &next_box_x);

		m = ' ';
		if (next_sok_y == (sok_y + 1)) m = 'd';
		if (next_sok_y == (sok_y - 1)) m = 'u';
		if (next_sok_x == (sok_x + 1)) m = 'r';
		if (next_sok_x == (sok_x - 1)) m = 'l';

		if (m == ' ')
			exit_with_error("move error");

		if ((current_index % index_num) != (next_index % index_num))
			m = m + 'A' - 'a';

		moves[moves_num++] = m;

		current_index = next_index;
		next_index = src[current_index];
	}

	free(src);
	return moves_num;
}

char sol_of_level[100000];
int sol_of_level_size;



void write_solution_header()
{
	FILE *fp;
	char filename[2000];

#ifdef WIN32
	sprintf(filename, "%s\\solutions.sok", global_dir);
#else
	sprintf(filename, "%s/solutions.sok", global_dir);
#endif

	fp = fopen(filename, "w");
	if (fp)
	{
		fprintf(fp, "%s solution file\n", solver_name);
		fprintf(fp, "===========================\n");
		fprintf(fp, "Collection: %s\n", global_level_set_name);
		fclose(fp);
	}
}

void save_level_to_solution_file(board b)
{
	FILE *fp;
	char filename[2000];

	if ((height == 0) || (width == 0)) return;

#ifdef WIN32
	sprintf(filename, "%s\\solutions.sok", global_dir);
#else
	sprintf(filename, "%s/solutions.sok", global_dir);
#endif

	fp = fopen(filename, "a");
	if (fp == NULL) return;

	fprintf(fp, "\n");

	if (strcmp(level_title, "None") == 0)
	{
		fprintf(fp, "Level %d\n", level_id);
	}
	else
	{
		fprintf(fp, "%s\n", level_title);
	}
	fprintf(fp, "\n");

	save_board_to_file(b, fp);
	fprintf(fp, "\n");

	fclose(fp);
}


int count_pushes()
{
	int i;
	int sum = 0;

	for (i = 0; i < sol_of_level_size; i++)
	{
		if (sol_of_level[i] == 'U') sum++;
		if (sol_of_level[i] == 'R') sum++;
		if (sol_of_level[i] == 'D') sum++;
		if (sol_of_level[i] == 'L') sum++;
	}
	return sum;
}

int global_total_pushes = 0;
int global_total_moves = 0;
int total_sol_time = 0;
int total_levels_tried = 0;
int total_levels_solved = 0;

void write_log_header()
{
	FILE *fp;
	char s[2000];
	char date_string[200];
	char time_string[200];


#ifdef WIN32
	sprintf(s, "%s\\times.txt", global_dir);
#else
	sprintf(s, "%s/times.txt", global_dir);
#endif

	get_date_as_string(time_string, date_string);

	fp = fopen(s, "w");
	if (fp == NULL)
		return;

	fprintf(fp, "--------------------------------------------------------------------------------------\n");
	fprintf(fp, "Sokoban Solver Statistics - Curry\n");
	fprintf(fp, "Date: %s  %s\n", date_string, time_string);
	fprintf(fp, "--------------------------------------------------------------------------------------\n");
	fprintf(fp, "   Task       Time Plugin               Result                     Moves  Pushes Level\n");
	fprintf(fp, "--------------------------------------------------------------------------------------\n");
	fclose(fp);

}

void save_level_log(int solved)
{
	FILE *fp;
	static int record_num = 1;
	int sol_time;
	char s[2000];

	sol_time = end_time - start_time;
	total_sol_time += sol_time;

	total_levels_tried++;
	if (solved)
		total_levels_solved++;

#ifdef WIN32
	sprintf(s, "%s\\times.txt", global_dir);
#else
	sprintf(s, "%s/times.txt", global_dir);
#endif


	fp = fopen(s, "a");
	if (fp == NULL)
		return;

	fprintf(fp, "%7d  ", record_num++);
	get_sol_time_as_hms(sol_time, s);

	fprintf(fp, " %s ", s);

	fprintf(fp, "%-21s", solver_name);

	if (solved)
	{
		fprintf(fp, "OK                     ");
	}
	else
	{
		sprintf(s, "%-23s", global_fail_reason);
		fprintf(fp, "%s", s);
	}

	if (solved)
	{
		fprintf(fp, "%9d", sol_of_level_size);
		fprintf(fp, "%8d ", count_pushes());
	}
	else
		fprintf(fp, "                  ");

	fprintf(fp, "%s\n", level_title);

	fclose(fp);

	if (solved)
	{
		global_total_moves += sol_of_level_size;
		global_total_pushes += count_pushes();
	}
}

void write_log_footer()
{
	FILE *fp;
	int failed,i;
	char s[2000];

	failed = total_levels_tried - total_levels_solved;
#ifdef WIN32
	sprintf(s, "%s\\times.txt", global_dir);
#else
	sprintf(s, "%s/times.txt", global_dir);
#endif

	fp = fopen(s, "a");
	if (fp == NULL)
		return;

	fprintf(fp, "--------------------------------------------------------------------------------------\n");
	fprintf(fp, ":%6d  ", total_levels_tried);

	get_sol_time_as_hms(total_sol_time, s);
	fprintf(fp, " %s ", s);

	sprintf(s, "Total                OK: %d Failed: %d      ", total_levels_solved, failed);
	fprintf(fp, "%s", s);

	for (i = strlen(s); i < 44; i++)
		fprintf(fp, " ");


	fprintf(fp, "%9d", global_total_moves);
	fprintf(fp, "%8d", global_total_pushes);
	fprintf(fp, "     :\n");
	fprintf(fp,"--------------------------------------------------------------------------------------\n");
	fclose(fp);
}

void save_times_to_solutions_file(int sol_time)
{
	char filename[1000];
	FILE *fp;
	char hms[200];
	char date[200];
	char time[200];

#ifdef WIN32
	sprintf(filename, "%s\\solutions.sok", global_dir);
#else
	sprintf(filename, "%s/solutions.sok", global_dir);
#endif

	fp = fopen(filename, "a");
	if (fp == 0) return;

	get_sol_time_as_hms(sol_time, hms);
	get_date_as_string(time, date);

	fprintf(fp, "\n");
	fprintf(fp, "Solver: %s", solver_name);
	fprintf(fp, "\n");

	fprintf(fp, "Solver time: %s\n", hms);
	fprintf(fp, "Solver date: %s  %s\n", date, time);
	fprintf(fp, "\n");
	fclose(fp);
}



void save_solution_from_moves(move *moves, int moves_num)
{
	move m;
	int i, n;
	board b;
	FILE *fp;
	char filename[2000];

	move move_played[MAX_SOL_LEN];

	int sok_y, sok_x, next_sok_y, next_sok_x;
	int start_box_index, end_box_index;
	int end_box_y, end_box_x;

	n = moves_num;

	if (n >= MAX_SOL_LEN) exit_with_error("solution is too long");
	if (n == 0) exit_with_error("no solution moves");

	for (i = 0; i < n; i++)
		move_played[i] = moves[i];


	sok_y = global_initial_player_y;
	sok_x = global_initial_player_x;

	copy_board(initial_board, b);
	sol_of_level_size = 0;

	for (i = 0; i < n; i++)
	{
		m = move_played[i];

		start_box_index = m.from;
		end_box_index = m.to;

		index_to_y_x(end_box_index, &end_box_y, &end_box_x);

		next_sok_y = end_box_y + delta_y[m.player_position];
		next_sok_x = end_box_x + delta_x[m.player_position];

		moves_num = find_sol_move(b, y_x_to_index(sok_y, sok_x), start_box_index,
			y_x_to_index(next_sok_y, next_sok_x), end_box_index, sol_of_level + sol_of_level_size);

		sol_of_level_size += moves_num;

		if (sol_of_level_size > 90000)
			exit_with_error("solution is too long");

		sok_y = next_sok_y;
		sok_x = next_sok_x;

		apply_move(b, &m);
	}

#ifdef WIN32
	sprintf(filename, "%s\\solutions.sok", global_dir);
#else
	sprintf(filename, "%s/solutions.sok", global_dir);
#endif

	fp = fopen(filename, "a");
	if (fp == 0) return;
	fprintf(fp, "Solution\n");
	for (i = 0; i < sol_of_level_size; i++)
		fprintf(fp, "%c", sol_of_level[i]);
	fprintf(fp, "\n");
	fclose(fp);
}

int solution_index(episodes *es)
{
	int i;

	for (i = 0; i < es->episodes_num; i++)
		if (es->e[i].box_num == global_boxes_in_level)
			return i;
	return -1;
}

int save_solution_if_found(episodes *es)
{
	int i;
	episode *sol;

	i = solution_index(es);
	if (i == -1) return 0;

	sol = es->e + i;
	print_episode(sol);
	save_solution_from_moves(sol->moves, sol->len);
	return 1;
}
