// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#ifndef __TREE
#define __TREE

#include "board.h"
#include "moves.h"

typedef struct move_hash_data
{
	move move;
	UINT_64 hash;
	char deadlocked;
	float weight;
} move_hash_data;


typedef struct node_element
{
	UINT_64 hash;
	int expansion;
	char deadlocked;
	char wins;
} node_element;

typedef struct expansion_data
{
	UINT_8 *b;
	node_element *node;
	expansion_data *father;
	int moves_num;
	int move_hash_place;
	int has_corral;
	int depth;
	int walk_id;
	int move_from_father;
	double value;

	double weight;
	double best_weight; // weight of best child
} expansion_data;



typedef struct tree
{
	// hash_array
	int *hash_array;
	int hash_size;
	int mask;

	// nodes
	node_element *nodes;
	int nodes_num;
	int max_nodes;

	// expansions
	expansion_data *expansions;
	int expansions_num;
	int max_expansions;

	// move-hash
	move_hash_data *move_hashes;
	int move_hashes_num;
	int max_move_hashes;

	// boards
	UINT_8 *boards;
	long boards_size;
} tree;

void init_tree(tree *t, int log_max_nodes);
int tree_nearly_full(tree *t, char *reason);
void set_root(tree *t, board b);
void expand_node(tree *t, expansion_data *e, int son);
void reset_tree(tree *t);
void print_node(tree *t, expansion_data *e, int print_moves);
int find_node_by_hash(tree *t, UINT_64 hash);
move get_move_to_node(tree *t, expansion_data *e);
int node_is_solved(tree *t, expansion_data *e);
void set_deadlock_status(tree *t, expansion_data *e);
expansion_data *last_expansion(tree *t);
int has_a_winning_son(tree *t, expansion_data *e, int *son);
node_element *get_son(tree *t, expansion_data *e, int son);
int son_is_expanded(tree *t, expansion_data *e, int son);
int son_is_deadlocked(tree *t, expansion_data *e, int son);
expansion_data* go_to_son(tree *t, expansion_data *e, int son);



#endif