// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>

#include "util.h"
#include "board.h"

#ifdef WIN32
#include <conio.h>
#else
#include <termios.h>
#include <unistd.h>
#endif

#ifdef WIN32
#include <conio.h>
#endif


void exit_with_error(const char *message)
{
	printf("%s", message);
	my_getch();
	exit(0);
}


void remove_newline(char *s)
{
	while (*s)
	{
		if (((*s) == '\r') || ((*s) == '\n'))
		{
			*s = 0;
			return;
		}
		s++;
	}
}

void remove_preceeding_spaces(char *s)
{
	int i, d, n;
	n = strlen(s);

	for (i = 0; i < n; i++)
		if (s[i] != ' ')
			break;
	d = i;

	for (i = 0; i < (n - d); i++)
		s[i] = s[i + d];
	s[i] = 0;
}





int get_number_at_end_of_string(char *s)
{
	int n = strlen(s);
	int i;

	for (i = n - 1; i >= 0; i--)
	{
		if ((s[i] < '0') || (s[i] > '9'))
			break;
	}

	if (i == (n - 1)) return -1;
	
	sscanf(s + i + 1, "%d", &n);
	return n;
}


#ifndef _WIN32
char linux_getch(void)
{
  // code from stackoverflow by mf_
    char buf = 0;
    struct termios old = {0};
    fflush(stdout);
    if(tcgetattr(0, &old) < 0)
        perror("tcsetattr()");
    old.c_lflag &= ~ICANON;
    old.c_lflag &= ~ECHO;
    old.c_cc[VMIN] = 1;
    old.c_cc[VTIME] = 0;
    if(tcsetattr(0, TCSANOW, &old) < 0)
        perror("tcsetattr ICANON");
    if(read(0, &buf, 1) < 0)
        perror("read()");
    old.c_lflag |= ICANON;
    old.c_lflag |= ECHO;
    if(tcsetattr(0, TCSADRAIN, &old) < 0)
        perror("tcsetattr ~ICANON");
    printf("%c\n", buf);
    return buf;
}
#endif

void my_getch()
{
#ifdef WIN32
         _getch();
#else
  linux_getch();
#endif
}





void get_date_as_string(char *time_out, char *date)
{
	time_t rawtime;
	struct tm * timeinfo;

	time(&rawtime);
	timeinfo = localtime(&rawtime);

	strftime(time_out, 80, "%X", timeinfo);
	strftime(date    , 80, "%Y-%m-%d", timeinfo);
}

void get_sol_time_as_hms(int sol_time, char *hms)
{
	sprintf(hms, "%02d:%02d:%02d", sol_time / 3600, (sol_time / 60) % 60, sol_time % 60);
}


void get_k_of_n(int k, int n, int *a)
{
	int i,p;
	int chosen = 0;

	if (n <= 0) exit_with_error("n<=0");
	if (k > n)  exit_with_error("k>n");

	for (i = 0; i < n; i++)
		a[i] = 0;

	while (chosen < k)
	{
		p = rand() % n;
		if (a[p] == 0)
		{
			a[p] = 1;
			chosen++;
		}
	}
}

void normalize_vector(float *a, int n)
{
	int i;
	float sum = 0;

	for (i = 0; i < n; i++)
		sum += a[i];

	if (sum == 0)
	{
		// choose uniformly in this case
		for (i = 0; i < n; i++)
			a[i] = 1 / (float)(n);
		return;
	}

	sum = 1 / sum;

	for (i = 0; i < n; i++)
		a[i] *= sum;
}

int choose_with_probability(float *a, int n)
{
	// this routine assumes that sum(a)=1
	int i;
	double sum = 0;
	double p;

	p = rand() / (double)RAND_MAX;
	if (p == 0) p += 0.01;
	if (p == 1) p -= 0.01;

	for (i = 0; i < n; i++)
	{
		sum += a[i];
		if (sum >= p)
			return i;
	}

	return n - 1; // shouldn't really be here, just in case
}